#include "datetime.hpp"
#include <system_error>
#include <stdexcept>

struct tm datetime_t::to_tm() const {
	struct tm tm;
	gmtime_r(&value.tv_sec, &tm);
	return tm;
}

datetime_t datetime_t::operator+(const interval_t & interval) const {
	switch (interval.unit) {
		case interval_t::seconds: return datetime_t(value.tv_sec + interval.amount, value.tv_usec);
		case interval_t::minutes: return datetime_t(value.tv_sec + (interval.amount * 60), value.tv_usec);
		case interval_t::hours: return datetime_t(value.tv_sec + (interval.amount * 60 * 60), value.tv_usec);
		case interval_t::days: return datetime_t(value.tv_sec + (interval.amount * 60 * 60 * 24), value.tv_usec);
		case interval_t::weeks: return datetime_t(value.tv_sec + (interval.amount * 60 * 60 * 24 * 7), value.tv_usec);
		default: return *this;
	}
}

datetime_t & datetime_t::set(
	const struct tm & tm,
	suseconds_t msec
) {
	auto copy = tm;
	value.tv_sec = timegm(&copy);
	value.tv_usec = msec;
	return *this;
}

datetime_t datetime_t::now() {
	struct timespec ts;
	if (clock_gettime(CLOCK_REALTIME, &ts) != 0) {
		throw std::system_error(errno, std::system_category());
	}
	return datetime_t(ts.tv_sec, ts.tv_nsec / 1000);
}

std::string datetime_t::format(const char * fmt) const {
	const auto tm = to_tm();
	std::string result(200, 0);
	result.resize(strftime(&result[0], result.size(), fmt, &tm));
	return result;
}
