#ifndef __DEDUCE_HPP__
#define __DEDUCE_HPP__

#include <tuple>

template<typename>
struct deduce_impl;

template<typename R, typename ...Args>
struct deduce_impl<R(Args...)> {
	using return_type = R;
	using tuple_type = std::tuple<Args...>;
};

template<typename R, typename ...Args>
struct deduce_impl<R(Args..., ...)> {
	using return_type = R;
	using tuple_type = std::tuple<Args...>;
};

template<typename R, typename ...Args>
struct deduce_impl<R(*)(Args...)> {
	using return_type = R;
	using tuple_type = std::tuple<Args...>;
};

template<typename R, typename ...Args>
struct deduce_impl<R(*)(Args..., ...)> {
	using return_type = R;
	using tuple_type = std::tuple<Args...>;
};

template<typename R, typename ...Args>
struct deduce_impl<R(&)(Args...)> {
	using return_type = R;
	using tuple_type = std::tuple<Args...>;
};

template<typename R, typename ...Args>
struct deduce_impl<R(&)(Args..., ...)> {
	using return_type = R;
	using tuple_type = std::tuple<Args...>;
};

template<typename R, typename T, typename ...Args>
struct deduce_impl<R(T::*)(Args...)> {
	using return_type = R;
	using class_type = T;
	using tuple_type = std::tuple<Args...>;
};

template<typename R, typename T, typename ...Args>
struct deduce_impl<R(T::*)(Args..., ...)> {
	using return_type = R;
	using class_type = T;
	using tuple_type = std::tuple<Args...>;
};

template<typename R, typename T, typename ...Args>
struct deduce_impl<R(T::*)(Args...) &> {
	using return_type = R;
	using class_type = T;
	using tuple_type = std::tuple<Args...>;
};

template<typename R, typename T, typename ...Args>
struct deduce_impl<R(T::*)(Args..., ...) &> {
	using return_type = R;
	using class_type = T;
	using tuple_type = std::tuple<Args...>;
};

template<typename R, typename T, typename ...Args>
struct deduce_impl<R(T::*)(Args...) &&> {
	using return_type = R;
	using class_type = T;
	using tuple_type = std::tuple<Args...>;
};

template<typename R, typename T, typename ...Args>
struct deduce_impl<R(T::*)(Args..., ...) &&> {
	using return_type = R;
	using class_type = T;
	using tuple_type = std::tuple<Args...>;
};

template<typename R, typename T, typename ...Args>
struct deduce_impl<R(T::*)(Args...) const> {
	using return_type = R;
	using class_type = T;
	using tuple_type = std::tuple<Args...>;
};

template<typename R, typename T, typename ...Args>
struct deduce_impl<R(T::*)(Args..., ...) const> {
	using return_type = R;
	using class_type = T;
	using tuple_type = std::tuple<Args...>;
};

template<typename R, typename T, typename ...Args>
struct deduce_impl<R(T::*)(Args...) const &> {
	using return_type = R;
	using class_type = T;
	using tuple_type = std::tuple<Args...>;
};

template<typename R, typename T, typename ...Args>
struct deduce_impl<R(T::*)(Args..., ...) const &> {
	using return_type = R;
	using class_type = T;
	using tuple_type = std::tuple<Args...>;
};

template<typename R, typename T, typename ...Args>
struct deduce_impl<R(T::*)(Args...) const &&> {
	using return_type = R;
	using class_type = T;
	using tuple_type = std::tuple<Args...>;
};

template<typename R, typename T, typename ...Args>
struct deduce_impl<R(T::*)(Args..., ...) const &&> {
	using return_type = R;
	using class_type = T;
	using tuple_type = std::tuple<Args...>;
};

template <typename T, typename = void>
struct deduce : deduce_impl<T> {};

template <typename T>
struct deduce<T, decltype(void(&T::operator()))> : deduce_impl<decltype(&T::operator())> {};

template<typename T>
using return_type_t = typename deduce<T>::return_type;

template<typename T>
using class_type_t = typename deduce<T>::class_type;

template<typename T>
using tuple_type_t = typename deduce<T>::tuple_type;

#endif
