#ifndef __INDEX_SEQUENCE_HPP__
#define __INDEX_SEQUENCE_HPP__

#include <cstddef>
#include <tuple>
#include <utility>
#include <type_traits>

template<typename T>
using remove_reference_t = typename std::remove_reference<T>::type;

#if (__cplusplus < 201402L)

template<typename T, T ...I>
struct integer_sequence {
	using value_type = T;
	static constexpr size_t size() noexcept { return sizeof...(I); }
};

namespace detail {

template<typename T, size_t N, T ...I>
struct make_integer_sequence : make_integer_sequence<T, N - 1, T(N - 1), I...> {};

template<typename T, T ...I>
struct make_integer_sequence<T, 0, I...> {
	using type = integer_sequence<T, I...>;
};

} // detail

template<size_t ...I>
using index_sequence = integer_sequence<size_t, I...>;

template<typename T, T N>
using make_integer_sequence = typename detail::make_integer_sequence<T, N>::type;

template<size_t N>
using make_index_sequence = make_integer_sequence<size_t, N>;

#else

template<size_t... I>
using index_sequence = std::integer_sequence<size_t, I...>;

template<size_t N>
using make_index_sequence = std::make_integer_sequence<size_t, N>;

#endif // C++11 polyfills

#if (__cplusplus < 201703L)

template <typename F, typename T, size_t ...I>
constexpr auto apply(F && fn, T && tuple, index_sequence<I...>)
	-> decltype(fn(std::get<I>(std::forward<T>(tuple))...)) {
	return fn(std::get<I>(std::forward<T>(tuple))...);
}

template<typename T>
using tuple_index_sequence = make_index_sequence<std::tuple_size<remove_reference_t<T>>::value>;

template <typename F, typename T>
constexpr auto apply(F && fn, T && tuple)
	-> decltype(apply(std::forward<F>(fn), std::forward<T>(tuple), tuple_index_sequence<T>{})) {
	return apply(std::forward<F>(fn), std::forward<T>(tuple), tuple_index_sequence<T>{});
}

#endif // C++14 polyfills

#endif
