#ifndef __UTILS_HPP__
#define __UTILS_HPP__

#include <cstdio>
#include <string>
#include <stdexcept>
#include <syslog.h>

void open_log();

template<typename ... arg_types>
void log(const char * format, arg_types && ... args) {
	open_log();
	syslog(LOG_INFO, format, std::forward<arg_types>(args) ...);
}

template<typename ... arg_types>
std::string format(const char * format, arg_types && ... args) {
	const auto len = std::snprintf(nullptr, 0, format, std::forward<arg_types>(args) ...);
	if (len < 0) {
		throw std::runtime_error("Format error");
	}
	std::string output(len, 0);
	std::snprintf(&output[0], len + 1, format, std::forward<arg_types>(args) ...);
	return output;
}

#endif
